module HooverTask (buildBeliefState, solveHooverTask) where

import Search.Heuristic
import qualified Data.Set as S
import qualified Data.Map as M

data HooverAction = GoUp | GoDown | GoLeft | GoRight | Suck
  deriving (Eq, Ord, Enum, Bounded, Show)

-- Физическое состояние агента
data PhysicalState = PhysicalState {
    hooverX :: Int, -- координата пылесоса X
    hooverY :: Int, -- координата пылесоса Y
    worldWidth :: Int, -- ширина мира
    worldHeight :: Int, -- высота мира
    wasCleanedUp :: M.Map (Int, Int) Bool -- какие клетки были очищены
  } deriving (Eq, Ord, Show)

buildPhysicalState :: Int -> Int -> Int -> Int -> Maybe PhysicalState
buildPhysicalState width height hx hy
  | width < 1 || height < 1 || hx < 0 || hy < 0 || hx >= width || hy >= height = Nothing
  | otherwise = Just physState
  where
    physState = PhysicalState { worldWidth = width, worldHeight = height, hooverX = hx, hooverY = hy, wasCleanedUp = cleanedUp }
    cleanedUp = foldl (\m k -> M.insert k False m) M.empty [(x, y) | x <- [0..width-1], y <- [0..height-1] ]

applyActionPhys :: PhysicalState -> HooverAction -> PhysicalState
applyActionPhys st GoUp = st { hooverX = min ( max 0 (hooverX st + 1) ) (worldWidth st - 1) }
applyActionPhys st GoRight = st { hooverY = min ( max 0 (hooverY st + 1) ) (worldHeight st - 1) }
applyActionPhys st GoDown = st { hooverX = min ( max 0 (hooverX st - 1) ) (worldWidth st - 1) }
applyActionPhys st GoLeft = st { hooverY = min ( max 0 (hooverY st - 1) ) (worldHeight st - 1) }
applyActionPhys st Suck = st { wasCleanedUp = M.insert (hooverX st, hooverY st) True (wasCleanedUp st) }

isCleanPhys :: PhysicalState -> Bool
isCleanPhys st = M.fold (\x y -> x && y) True (wasCleanedUp st)

-- минимальное количество шагов, необходимое для очистки пола
hooverHeuristicPhys :: PhysicalState -> Int
hooverHeuristicPhys st = 2 * dirtyCells - 1 -- не учитываем текущее положение пылесоса
  where
    dirtyCells = M.fold (\x y -> if x then y else y + 1) 0 (wasCleanedUp st)

-- Логическое "И" над двумя физическими состояниями
physicalStateConjunction :: PhysicalState -> PhysicalState -> PhysicalState
physicalStateConjunction a b = PhysicalState {
    hooverX = 0, hooverY = 0,
    worldWidth = newWidth, worldHeight = newHeight,
    wasCleanedUp = foldl (\m k -> M.insert k ((wasCleanedUp a M.! k) && (wasCleanedUp b M.! k)) m) M.empty
      [(x, y) | x <- [0..newWidth-1], y <- [0..newHeight-1] ]
  }
  where
    newWidth = min (worldWidth a) (worldWidth b)
    newHeight = min (worldHeight a) (worldHeight b)

-- Доверительное состояние агента
data BeliefState = BeliefState {
    actionsLog :: [HooverAction], -- история действий
    physicalSet :: S.Set PhysicalState -- множество физических состояний
  } deriving (Show)

instance Eq BeliefState where
  x == y = physicalSet x == physicalSet y

instance Ord BeliefState where
  x `compare` y = physicalSet x `compare` physicalSet y

buildBeliefState :: Int -> Int -> Maybe BeliefState
buildBeliefState width height 
   | width < 1 || height < 1 = Nothing
   | otherwise = case physStates of
     Just set -> Just BeliefState { actionsLog = [], physicalSet = set }
     otherwise -> Nothing
   where
     physStates = foldl (\s (x, y) -> myInsert (buildPhysicalState width height x y) s) (Just S.empty) [(x, y) | x <- [0..width-1], y <- [0..height-1]]
     myInsert _ Nothing = Nothing
     myInsert Nothing _ = Nothing
     myInsert (Just i) (Just s) = Just (S.insert i s)

applyActionBelief :: BeliefState -> HooverAction -> BeliefState
applyActionBelief st action = st {
    actionsLog = actionsLog st ++ [action],
    physicalSet = S.map (\x -> applyActionPhys x action) (physicalSet st)
  }

isCleanBelief :: BeliefState -> Bool
isCleanBelief st = S.fold (\x y -> isCleanPhys x && y) True (physicalSet st)

hooverHeuristic :: BeliefState -> [BeliefState] -> Int
hooverHeuristic st xs =
  (length xs) + hooverHeuristicPhys (S.fold (\x y -> physicalStateConjunction x y) (S.findMin $ physicalSet st) (physicalSet st))

instance TaskState BeliefState where
  nearbyStates st = 
    map (\action -> applyActionBelief st action) [minBound :: HooverAction .. maxBound :: HooverAction]

solveHooverTask st = case st of
  Just startState -> case graphSearch startState isCleanBelief hooverHeuristic of
    [] -> []
    lst -> show $ actionsLog (last lst)
  Nothing -> []
