module Search.Heuristic (
    TaskState,
    nearbyStates,
    breadthFirstSearch,
    depthFirstSearch,
    graphSearch
  ) where

import Search.Heuristic.PQueue as Q
import Data.Set as S -- Data.HashSet => TaskState should be Hashable (and not Ord?)
import Data.List as L

class (Eq a, Ord a) => TaskState a where
  nearbyStates :: a -> [a]

data TaskStateAndPath a =
  TaskStateAndPath { state :: a, path :: [a] }

instance Eq a => Eq (TaskStateAndPath a) where
  x == y = state x == state y

instance Ord a => Ord (TaskStateAndPath a) where
  compare x y = compare (state x) (state y)

instance Show a => Show (TaskStateAndPath a) where
  show x = show $ path x

stateAndPath st p = TaskStateAndPath { state = st, path = p }

-- Поиск в ширину
breadthFirstSearch :: TaskState a => a -> (a -> Bool) -> [a]
breadthFirstSearch x isGoal
  | isGoal x = [x]
  | otherwise =
      genericSearch [[x]] [] isGoal 
        (\open newOpen -> open ++ newOpen)

-- Поиск в глубину
depthFirstSearch :: TaskState a => a -> (a -> Bool) -> [a]
depthFirstSearch x isGoal
  | isGoal x = [x]
  | otherwise =
      genericSearch [[x]] [] isGoal
        (\open newOpen -> newOpen ++ open)

genericSearch :: TaskState a => [[a]] -> [a] -> (a -> Bool) -> ([[a]] -> [[a]] -> [[a]]) -> [a]
genericSearch [] _ _ _ = [] -- если список open пуст, решения нет
genericSearch open closed isGoal modifyOpen
  | length solution > 0 = reverse $ head solution
  | otherwise =
    genericSearch (modifyOpen pathList newOpen)
      newClosed isGoal modifyOpen
    where
      currPath@(currState:_) = head open
      pathList = tail open
      newClosed = (currState:closed)
      -- если точно нет повторов, можно убрать nub
      newOpen = L.map (\x -> (x:currPath))
        (((nub.nearbyStates) currState) L.\\
          (newClosed ++ L.map head pathList))
      -- ищем решение во всех новых элементах open
      -- это дает существенный прирост скорости
      solution = take 1
        $ L.filter (\(st:_) -> isGoal st) newOpen

-- Упрощенный поиск для алгоритма A*
graphSearch :: (TaskState s) => s -> (s -> Bool) -> (s -> [s] -> Int) -> [s]
graphSearch st isGoal heuristicFunc = 
  graphSearch' isGoal heuristicFunc 
    -- (Q.singleton (stateAndPath st [st]) 0 :: ??? ) 
    (Q.insert (stateAndPath st [st]) 0 q) 
    (S.empty)
  where
    q = Q.empty :: (Ord x) => PQueue (TaskStateAndPath x) Int

graphSearch' isGoal func open closed =
  case Q.findMin open of
  Nothing -> []
  Just found ->
    if isGoal.state.fst $ found
      then reverse.path.fst $ found
      else nextStep (state.fst $ found) (path.fst $ found)
  where
    nextStep st p = graphSearch' isGoal func newOpen newClosed
      where
      newClosed = S.insert st closed
      newOpen = L.foldl 
        (\q s -> insertState s p func q) ( Q.deleteMin open )
        (L.filter (\x -> not $ S.member x newClosed)
          $ (nub.nearbyStates) st)

insertState s p func q =
  case Q.lookup stp q of
  Nothing -> ins q
  Just f ->
    if func s p >= f
      then q 
      else ins $ Q.delete stp q
  where
    ins = Q.insert stp (func s p)
    stp = stateAndPath s (s:p)

